/*
    Copyright (c) 2005-2024 Leisenfels GmbH. All rights reserved.
    Use is subject to license terms.
*/

package com.lf.vfslib;

// Do not import provider specific classes here, must be usable separately (NoClassDefFoundError)

import com.lf.vfslib.core.VFSLibSettings;
import com.lf.vfslib.dropbox.DbxFileProvider;
import com.lf.vfslib.gdrive.GDriveFileProvider;
import com.lf.vfslib.lang.ReflectionUtils;
import com.lf.vfslib.s3.S3FileProvider;
import org.apache.commons.vfs2.impl.DefaultFileSystemManager;

import java.util.Arrays;
import java.util.logging.Level;


/**
 * Configures and activates the VFSLib providers to access network file systems via Commons VFS.
 * <p>
 * Use the shared instance if you have only one single file system manager instance. Create multiple
 * instances with different configurations with the default constructor methods if there are
 * separate file system managers in use.
 *
 * @author Axel Schwolow
 * @created 2016-01-01
 * @since 1.6
 */
public final class VFSLib {


    /**
     * The shared <code>VFSLib</code> instance.
     */
    private static VFSLib sharedInstance = null;
    /**
     * The default scheme name for Dropbox "dropbox", override with "dbx" or whatever.
     */
    private String schemeDropbox = "dropbox";
    /**
     * The default scheme name for Google Drive "gdrive", override with "google" or whatever.
     */
    private String schemeGoogleDrive = "gdrive";
    /**
     * The default scheme name for Amazon S3 "s3", override with "amazon" or whatever.
     */
    private String schemeAmazonS3 = "s3";


    /**
     * Constructor method.
     *
     * @since 1.6
     */
    public VFSLib() {
    }

    /**
     * Provides the shared instance of the <code>VFSLib</code>.
     *
     * @return Shared instance
     * @since 1.6
     */

    public static synchronized VFSLib getSharedInstance() {

        if (sharedInstance == null) sharedInstance = new VFSLib();
        return sharedInstance;
    }

    /**
     * Setter method for the VFSLib Dropbox scheme.
     * <p>
     * The default scheme name for Dropbox is "dropbox", override with "dbx" or whatever.
     * This method must be called before the providers are being installed.
     *
     * @param scheme The scheme
     * @throws NullPointerException     If the scheme is <code>null</code> or empty
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public void setSchemeDropbox(String scheme) {

        if (scheme == null) throw new NullPointerException();
        if (scheme.trim().isEmpty()) {
            VFSLibSettings.logMessage("VFSLIB-0013", Level.WARNING);
            throw new IllegalArgumentException(VFSLibSettings.getUserMessage("VFSLIB-0013"));
        }
        this.schemeDropbox = scheme;
    }

    /**
     * Getter method for the VFSLib Dropbox scheme.
     *
     * @return The scheme
     * @since 1.6
     */
    public String getSchemeDropbox() {
        return this.schemeDropbox;
    }

    /**
     * Configures the given file system manager to use the VFSLib provider for Dropbox (convenience method).
     * <p>
     * The required libraries must be in the classpath.
     * Please setup the application info and request configuration for Dropbox first. You may
     * use a different scheme/protocol name by calling <code>setSchemeDropbox()</code> (default is "dropbox").
     *
     * @param manager          The VFSLib manager (required)
     * @param appkey           The Dropbox application info (required for <code>DbxAppInfo</code>)
     * @param appsecret        The Dropbox request configuration (required for <code>DbxAppInfo</code>)
     * @param clientidentifier The Dropbox client identifier (required for <code>DbxRequestConfig</code>)
     * @return The installed provider, <code>null</code> if an error occurred
     * @throws NullPointerException     If a parameter is <code>null</code>
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public DbxFileProvider addProviderDropbox(DefaultFileSystemManager manager, String appkey, String appsecret, String clientidentifier) {

        if (manager == null) throw new NullPointerException();
        if (appkey == null) throw new NullPointerException();
        if (appsecret == null) throw new NullPointerException();
        if (clientidentifier == null) throw new NullPointerException();

        if (!checkLibsDropbox()) return null;

        // Create Dropbox instances
        //DbxAppInfo appinfo = new DbxAppInfo(appkey, appsecret);
        //DbxRequestConfig reqconfig = new DbxRequestConfig(clientidentifier);

        // Avoid imports, use reflection here
        boolean silent = true;  // Set to true for production builds!
        Object appinfo = ReflectionUtils.newInstance(silent, "com.dropbox.core.DbxAppInfo", new Class[]{String.class, String.class}, appkey, appsecret);
        Object reqconfig = ReflectionUtils.newInstance(silent, "com.dropbox.core.DbxRequestConfig", new Class[]{String.class}, clientidentifier);

        return this.addProviderDropbox(manager, appinfo, reqconfig);
    }

    /**
     * Configures the given file system manager to use the VFSLib provider for Dropbox.
     * <p>
     * The required libraries must be in the classpath.
     * Please setup the application info and request configuration for Dropbox first. You may
     * use a different scheme/protocol name by calling <code>setSchemeDropbox()</code> (default is "dropbox").
     *
     * @param manager   The VFSLib manager (required)
     * @param appinfo   The Dropbox application info (required, type is <code>com.dropbox.core.DbxAppInfo</code>)
     * @param reqconfig The Dropbox request configuration (required, type is <code>com.dropbox.core.DbxRequestConfig</code>)
     * @return The installed provider, <code>null</code> if an error occurred
     * @throws NullPointerException     If a parameter is <code>null</code>
     * @throws IllegalArgumentException If a parameter is not of the expected type
     * @since 1.6
     */
    public DbxFileProvider addProviderDropbox(DefaultFileSystemManager manager, Object appinfo, Object reqconfig) {

        if (manager == null) throw new NullPointerException();
        if (appinfo == null) throw new NullPointerException();
        if (reqconfig == null) throw new NullPointerException();

        if (!checkLibsDropbox()) return null;

        boolean ok = false;
        DbxFileProvider provider = null;
        try {
            provider = new DbxFileProvider(appinfo, reqconfig, this);
            manager.addProvider(this.schemeDropbox, provider);
            ok = true;
        } catch (Exception e) {
            VFSLibSettings.logMessage("VFSLIB-0017", Level.WARNING);
            System.err.println(VFSLibSettings.getUserMessage("VFSLIB-0017"));
        }
        return ok ? provider : null;
    }

    /**
     * Checks if the required libraries are accessible.
     *
     * @return Available?
     * @since 1.6
     */
    private static boolean checkLibsDropbox() {

        try {  // Throws
            Class.forName("com.dropbox.core.DbxSdkVersion");       // dropbox-core-sdk-1.7.6.jar
            Class.forName("com.fasterxml.jackson.core.Version");   // jackson-core-2.11.1.jar
            return true;
        } catch (Exception e) {
            VFSLibSettings.logMessage("VFSLIB-0016", Level.WARNING);
        }
        return false;
    }

    /**
     * Setter method for the VFSLib Google Drive scheme.
     * <p>
     * The default scheme name for Google Drive is "gdrive", override with "google" or whatever.
     * This method must be called before the providers are being installed.
     *
     * @param scheme The scheme
     * @throws NullPointerException     If the scheme is <code>null</code> or empty
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public void setSchemeGoogleDrive(String scheme) {

        if (scheme == null) throw new NullPointerException();
        if (scheme.trim().isEmpty()) {
            VFSLibSettings.logMessage("VFSLIB-0018", Level.WARNING);
            throw new IllegalArgumentException(VFSLibSettings.getUserMessage("VFSLIB-0018"));
        }
        this.schemeGoogleDrive = scheme;
    }

    /**
     * Getter method for the VFSLib Google Drive scheme.
     *
     * @return The scheme
     * @since 1.6
     */
    public String getSchemeGoogleDrive() {
        return this.schemeGoogleDrive;
    }

    /**
     * Configures the given file system manager to use the VFSLib provider for Google Drive (convenience method).
     * <p>
     * The required libraries must be in the classpath.
     * Please setup a project and application in the Google Cloud Console first. Don't forget to
     * have the Drive API enabled for your project registered from within the Cloud Console.
     * <p>
     * You may use a different scheme/protocol name by calling <code>getSchemeGoogleDrive()</code>
     * (default is "gdrive").
     *
     * @param manager      The VFSLib manager (required)
     * @param clientid     The Google Drive client ID (required)
     * @param clientsecret The Google Drive client secret (required)
     * @param appname      The application name (required)
     * @param redirecturi  The Google Drive redirect URI (required)
     * @return The installed provider, <code>null</code> if an error occurred
     * @throws NullPointerException     If a parameter is <code>null</code>
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public GDriveFileProvider addProviderGoogleDrive(DefaultFileSystemManager manager, String clientid, String clientsecret, String appname, String redirecturi) {

        if (manager == null) throw new NullPointerException();
        if (clientid == null) throw new NullPointerException();
        if (clientsecret == null) throw new NullPointerException();
        if (appname == null) throw new NullPointerException();
        if (redirecturi == null) throw new NullPointerException();

        if (!checkLibsGoogleDrive()) return null;

        // Create Google Drive instances
        //HttpTransport httptransport = new NetHttpTransport();
        //JsonFactory jsonfactory = new JacksonFactory();

        // See https://developers.google.com/drive/quickstart-java
        /*
        GoogleAuthorizationCodeFlow authflow = new GoogleAuthorizationCodeFlow.Builder(
                httptransport, jsonfactory, clientid, clientsecret, Arrays.asList(DriveScopes.DRIVE))
                .setAccessType("online")
                .setApprovalPrompt("auto").build();
                */

        // Avoid imports, use reflection here
        boolean silent = true;  // Set to true for production builds!
        Object httptransport = ReflectionUtils.newInstance(silent, "com.google.api.client.http.javanet.NetHttpTransport", null);
        Object jsonfactory = ReflectionUtils.newInstance(silent, "com.google.api.client.json.jackson2.JacksonFactory", null);

        String[] array = new String[]{"https://www.googleapis.com/auth/drive"};  // May change!
        Object builder = ReflectionUtils.newInstance(//
                "com.google.api.client.googleapis.auth.oauth2.GoogleAuthorizationCodeFlow$Builder", silent, new String[]{"com.google.api.client.http.HttpTransport", "com.google.api.client.json.JsonFactory", //
                        "java.lang.String", "java.lang.String", "java.util.Collection"}, httptransport, jsonfactory, clientid, clientsecret, Arrays.asList(array));
        // public GoogleAuthorizationCodeFlow(com.google.api.client.http.HttpTransport transport,
        //                                    com.google.api.client.json.JsonFactory jsonFactory,
        //                                    java.lang.String clientId, java.lang.String clientSecret,
        //                                    java.util.Collection<java.lang.String> scopes) {...

        ReflectionUtils.invoke(silent, builder, "setAccessType", new Class[]{String.class}, "online");
        ReflectionUtils.invoke(silent, builder, "setApprovalPrompt", new Class[]{String.class}, "auto");
        Object authflow = ReflectionUtils.invoke(silent, builder, "build", null);

        return this.addProviderGoogleDrive(manager, authflow, appname, redirecturi);
    }

    /**
     * Configures the given file system manager to use the VFSLib provider for Google Drive.
     * <p>
     * The required libraries must be in the classpath.
     * Please setup a project and application in the Google Cloud Console first. Don't forget to
     * have the Drive API enabled for your project registered from within the Cloud Console.
     * <p>
     * You may use a different scheme/protocol name by calling <code>getSchemeGoogleDrive()</code>
     * (default is "gdrive").
     *
     * @param manager     The VFSLib manager (required)
     * @param authflow    The Google authorization flow (required, type is <code>com.google.api.client.googleapis.auth.oauth2.GoogleAuthorizationCodeFlow</code>)
     * @param appname     The application name (required)
     * @param redirecturi The Google Drive redirect URI (required)
     * @return The installed provider, <code>null</code> if an error occurred
     * @throws NullPointerException     If a parameter is <code>null</code>
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public GDriveFileProvider addProviderGoogleDrive(DefaultFileSystemManager manager, Object authflow, String appname, String redirecturi) {

        if (manager == null) throw new NullPointerException();
        if (authflow == null) throw new NullPointerException();
        if (appname == null) throw new NullPointerException();
        if (redirecturi == null) throw new NullPointerException();

        if (!checkLibsGoogleDrive()) return null;

        boolean ok = false;
        GDriveFileProvider provider = null;
        try {
            provider = new GDriveFileProvider(authflow, appname, redirecturi, this);
            manager.addProvider(this.schemeGoogleDrive, provider);
            ok = true;
        } catch (Exception e) {
            VFSLibSettings.logMessage("VFSLIB-0019", Level.WARNING);
            System.err.println(VFSLibSettings.getUserMessage("VFSLIB-0019"));
        }
        return ok ? provider : null;
    }

    /**
     * Checks if the required libraries are accessible.
     *
     * @return Available?
     * @since 1.6
     */
    private static boolean checkLibsGoogleDrive() {

        try {  // Throws
            Class.forName("com.google.api.client.googleapis.GoogleUtils");               // google-api-client-1.17.0-rc.jar
            Class.forName("com.google.api.services.drive.Drive");                        // google-api-services-drive-v2-rev102-1.17.0-rc.jar
            Class.forName("com.google.api.client.http.EmptyContent");                    // google-http-client-1.17.0-rc.jar
            Class.forName("com.google.api.client.json.gson.GsonFactory");                // google-http-client-gson-1.17.0-rc.jar
            Class.forName("com.google.api.client.json.jackson2.JacksonFactory");         // google-http-client-jackson2-1.17.0-rc.jar
            Class.forName("com.google.api.client.extensions.jdo.JdoDataStoreFactory");   // google-http-client-jdo-1.17.0-rc.jar
            Class.forName("com.google.api.client.auth.oauth.OAuthSigner");               // google-oauth-client-1.17.0-rc.jar
            Class.forName("javax.annotation.CheckForNull");                              // jsr305-1.3.9.jar
            return true;
        } catch (Exception e) {
            /*e.printStackTrace();*/
            VFSLibSettings.logMessage("VFSLIB-0020", Level.WARNING);
        }
        return false;
    }

    /**
     * Setter method for the VFSLib Amazon S3 scheme.
     * <p>
     * The default scheme name for Amazon S3 is "s3", override with "amazon" or whatever.
     * This method must be called before the providers are being installed.
     *
     * @param scheme The scheme
     * @throws NullPointerException     If the scheme is <code>null</code> or empty
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public void setSchemeAmazonS3(String scheme) {

        if (scheme == null) throw new NullPointerException();
        if (scheme.trim().isEmpty()) {
            VFSLibSettings.logMessage("VFSLIB-0022", Level.WARNING);
            throw new IllegalArgumentException(VFSLibSettings.getUserMessage("VFSLIB-0022"));
        }
        this.schemeAmazonS3 = scheme;
    }

    /**
     * Getter method for the VFSLib Amazon S3 scheme.
     *
     * @return The scheme
     * @since 1.6
     */
    public String getSchemeAmazonS3() {
        return this.schemeAmazonS3;
    }

    /**
     * Configures the given file system manager to use the VFSLib provider for Amazon S3 (convenience method).
     * <p>
     * The required libraries must be in the classpath.
     * Please setup a valid Amazon Web Services (AWS) account first. You may
     * use a different scheme/protocol name by calling <code>setSchemeAmazonS3()</code> (default is "s3").
     *
     * @param manager The VFSLib manager (required)
     * @return The installed provider, <code>null</code> if an error occurred
     * @throws NullPointerException     If a parameter is <code>null</code>
     * @throws IllegalArgumentException If an error occurs
     * @since 1.6
     */
    public S3FileProvider addProviderAmazonS3(DefaultFileSystemManager manager) {

        if (manager == null) throw new NullPointerException();

        if (!checkLibsAmazonS3()) return null;

        boolean ok = false;
        S3FileProvider provider = null;
        try {
            provider = new S3FileProvider(this);
            manager.addProvider(this.schemeAmazonS3, provider);
            ok = true;
        } catch (Exception e) {
            VFSLibSettings.logMessage("VFSLIB-0023", Level.WARNING);
            System.err.println(VFSLibSettings.getUserMessage("VFSLIB-0023"));
        }
        return ok ? provider : null;
    }

    /**
     * Checks if the required libraries are accessible.
     *
     * @return Available?
     * @since 1.6
     */
    private static boolean checkLibsAmazonS3() {

        try {  // Throws
            Class.forName("com.amazonaws.services.s3.iterable.S3Versions");       // aws-java-sdk-1.6.1.jar
            return true;
        } catch (Exception e) {
            VFSLibSettings.logMessage("VFSLIB-0024", Level.WARNING);
        }
        return false;
    }
}
