/*
    Copyright (c) 2005-2024 Leisenfels GmbH. All rights reserved.
    Use is subject to license terms.
*/

package com.lf.vfslib.dropbox;

import com.dropbox.core.DbxDownloader;
import com.dropbox.core.DbxException;
import com.dropbox.core.DbxUploader;
import com.dropbox.core.v2.files.*;
import com.lf.vfslib.VFSLib;
import com.lf.vfslib.core.VFSLibSettings;
import org.apache.commons.vfs2.*;
import org.apache.commons.vfs2.provider.AbstractFileName;
import org.apache.commons.vfs2.provider.AbstractFileObject;
import org.apache.commons.vfs2.provider.UriParser;
import org.apache.commons.vfs2.util.MonitorInputStream;
import org.apache.commons.vfs2.util.MonitorOutputStream;
import org.apache.commons.vfs2.util.RandomAccessMode;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * Represents a Dropbox file or folder.
 *
 * @author Axel Schwolow
 * @created 2016-01-01
 * @since 1.6
 */
public class DbxFileObject extends AbstractFileObject {

    /**
     * The underlying file system.
     */
    final protected DbxFileSystem fileSystem;
    /**
     * The relative path on the Dropbox server, the root / is not supported by Dropbox API v2, use "" instead.
     */
    protected String relPath = null;
    /**
     * Helper to avoid multiple refreshing.
     */
    protected boolean inRefresh = false;
    /**
     * The metadata for the underlying Dropbox entry, either <code>FileMetadata</code> or <code>FolderMetadata</code>.
     */
    protected Metadata dbxEntryMetadata = null;
    /**
     * The attributes specific for cloud file systems (e.g. "Content-Length" for proper upload).
     */
    protected Map<String, Object> attributes = null;


    /**
     * Constructor method.
     *
     * @param name       The file name
     * @param filesystem The file system
     * @throws NullPointerException If a parameter is <code>null</code>
     * @throws FileSystemException  If something goes wrong
     * @since 1.6
     */
    public DbxFileObject(AbstractFileName name, DbxFileSystem filesystem) throws FileSystemException {

        super(name, filesystem);

        this.fileSystem = filesystem;
        this.relPath = UriParser.decode(filesystem.getRootName().getRelativeName(name));
        if (this.relPath.equals(".") || this.relPath.equals("/")) this.relPath = "";  // Does not work with Dropbox!
        else if (!this.relPath.startsWith("/")) this.relPath = '/' + this.relPath;  // Files
    }

    /**
     * Clean-up method to help the gc.
     *
     * @throws Throwable Error indication
     * @since 1.6
     */
    @Override
    protected void finalize() throws Throwable {

        super.finalize();

        if (this.attributes != null) this.attributes.clear();
    }

    /**
     * Detach from the file system.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doDetach() throws Exception {
    }

    /**
     * Refresh this file.
     *
     * @throws FileSystemException If something goes wrong
     * @since 1.6
     */
    @Override
    public void refresh() throws FileSystemException {

        if (!this.inRefresh) {  // Avoid redundant refreshing
            try {
                this.inRefresh = true;
                super.refresh();
            } finally {
                this.inRefresh = false;
            }
        }
    }

    /**
     * Determines the type of this file, returns <code>null</code> if the file does not exist.
     *
     * @return The file type
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected FileType doGetType() throws Exception {

        // Possible VFS values:

        // FileType.FILE
        // FileType.FILE_OR_FOLDER
        // FileType.FOLDER
        // FileType.IMAGINARY

        statSelf();

        // May still be null if an anticipated file is queried (e.g. abc.txt.sha)
        if ("".equals(this.relPath)) return FileType.FOLDER; // The root / is not supported by Dropbox API v2
        if (this.dbxEntryMetadata != null) {
            if (this.dbxEntryMetadata instanceof FileMetadata) return FileType.FILE;
            else if (this.dbxEntryMetadata instanceof FolderMetadata) return FileType.FOLDER;
        }

        FileType type = FileType.IMAGINARY;
        try {
            // This is special here: IMAGINARY has no attributes, allow attributes for setting upload "Content-Length"
            Field field = type.getClass().getDeclaredField("hasAttrs");  // enum
            field.setAccessible(true);  // Override access limitations for private, protected etc.
            field.set(type, true);
        } catch (Exception ignored) {
        }
        return type;
    }

    /**
     * Determines the type of this file.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    protected void statSelf() throws Exception {

        if (this.dbxEntryMetadata != null) return;
        if ("".equals(this.relPath)) return; // The root / is not supported by Dropbox API v2, use "" instead
        //Thread.dumpStack();

        DbxClientWrapper client = this.fileSystem.getClient();
        try {
            this.dbxEntryMetadata = client.dbxClient.files().getMetadata(this.relPath);
        } catch (Exception e) {
            this.dbxEntryMetadata = null;  // Does not exist
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Called when the type or content of this file changes.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void onChange() throws Exception {

        this.dbxEntryMetadata = null;
        statSelf();
    }

    /**
     * Creates this file as a folder.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doCreateFolder() throws Exception {

        this.statSelf();

        DbxClientWrapper client = this.fileSystem.getClient();
        try {
            client.dbxClient.files().createFolderV2(this.relPath);
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Gets the last modification date and time.
     *
     * @return The last modification time in milliseconds
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected long doGetLastModifiedTime() throws Exception {

        this.statSelf();

        if (this.dbxEntryMetadata instanceof FileMetadata) { // Not available for folders
            try {
                Date modtime = ((FileMetadata) this.dbxEntryMetadata).getServerModified();
                if (modtime != null) return modtime.getTime();
            } catch (Exception ignored) {
            }
        }
        throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_UNKNOWN_MOD_TIME_ERROR"));
    }

    /**
     * Sets the last modified time of this file (not supported by Dropbox).
     *
     * @param modtime Last modification time in milliseconds
     * @return Successful?
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected boolean doSetLastModifiedTime(long modtime) throws Exception {
        return false;  // Not supported
    }

    /**
     * Deletes the file.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doDelete() throws Exception {

        this.statSelf();

        DbxClientWrapper client = this.fileSystem.getClient();
        try {
            client.dbxClient.files().deleteV2(this.relPath);
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Rename the file.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doRename(FileObject newfile) throws Exception {

        this.statSelf();

        DbxClientWrapper client = this.fileSystem.getClient();
        try {
            client.dbxClient.files().moveV2(this.relPath, ((DbxFileObject) newfile).relPath);
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Lists the children of this file.
     *
     * @return The list of children, may be empty
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected FileObject[] doListChildrenResolved() throws Exception {

        FileObject[] array = null;
        DbxClientWrapper client = this.fileSystem.getClient();
        FileSystemManager fsmanager = getFileSystem().getFileSystemManager();

        statSelf();
        try {
            ListFolderResult result = client.dbxClient.files().listFolder(this.relPath);
            List<Metadata> listing = result.getEntries();

            int size = listing.size();
            array = new FileObject[size];
            for (int i = 0; i < size; i++) {
                Metadata child = listing.get(i);
                array[i] = this.fileSystem.resolveFile(
                        fsmanager.resolveName(getName(), UriParser.encode(child.getPathLower()), NameScope.CHILD));
            }
        } finally {
            this.fileSystem.putClient(client);
        }
        return array;
    }

    /**
     * Lists the children of this file.
     *
     * @return The list of children, may be empty
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected String[] doListChildren() throws Exception {

        FileObject[] childs = this.doListChildrenResolved();
        String[] result = new String[childs.length];
        for (int i = 0; i < childs.length; i++) {
            result[i] = String.valueOf(childs[i]);
        }
        return result;
    }

    /**
     * Returns the size of the file content (in bytes).
     *
     * @return The size, -1 if not available
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected long doGetContentSize() throws Exception {

        statSelf();

        if (this.dbxEntryMetadata instanceof FileMetadata) { // Not available for folders
            try {
                return ((FileMetadata) this.dbxEntryMetadata).getSize();
            } catch (Exception e) {
                throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_UNKNOWN_SIZE_ERROR"));
            }
        }
        return -1L;
    }

    /**
     * Provides the instance to model random access for files.
     *
     * @return The instance
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected RandomAccessContent doGetRandomAccessContent(RandomAccessMode mode) throws Exception {
        return new DbxRandomAccessContent(this, mode);
    }

    /**
     * Creates an input stream to read the file content from.
     *
     * @return The input stream
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected InputStream doGetInputStream() throws Exception {

        this.statSelf();

        // Only makes sense for files...
        if (!getType().hasContent()) {
            String text = VFSLibSettings.getUserText(DbxClientWrapper.class + "_READ_NOT_FILE_ERROR");
            throw new FileSystemException(text, getName());
        }

        DbxClientWrapper client = this.fileSystem.getClient();

        synchronized (this.fileSystem) {
            try {
                DbxDownloader<FileMetadata> downloader = client.dbxClient.files().download(this.relPath, null);  // Latest revision
                return new DbxInputStream(client, downloader, downloader.getInputStream());
            } finally {
            }  // For client release and stream close see DbxInputStream
        }
    }

    /**
     * Creates an output stream to write the file content to.
     *
     * @param append Not supported currently, always created freshly
     * @return The output stream
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected OutputStream doGetOutputStream(boolean append) throws Exception {

        this.statSelf();

        // Only makes sense for files...
        //if (!getType().hasContent()) throw new FileSystemException("vfs.provider/write-not-file.error", getName());

        DbxClientWrapper client = this.fileSystem.getClient();

        synchronized (this.fileSystem) {
            DbxUploader<FileMetadata, UploadError, UploadErrorException> uploader = client.dbxClient.files().upload(this.relPath);
            return new DbxOutputStream(client, uploader, uploader.getOutputStream());
        }
    }

    /**
     * An <code>InputStream</code> that monitors for end-of-file.
     *
     * @since 1.6
     */
    private class DbxInputStream extends MonitorInputStream {

        private DbxClientWrapper client = null;
        private DbxDownloader<FileMetadata> downloader = null;

        public DbxInputStream(DbxClientWrapper client, DbxDownloader<FileMetadata> downloader, InputStream in) {

            super(in);
            this.client = client;
            this.downloader = downloader;
        }

        @Override
        protected void onClose() throws IOException {
            try {
                this.downloader.close();
            } finally {
                fileSystem.putClient(this.client);
            }
        }
    }

    /**
     * An <code>OutputStream</code> that wraps a Dropbox <code>OutputStream</code>, and closes it when the stream is closed.
     *
     * @since 1.6
     */
    private class DbxOutputStream extends MonitorOutputStream {


        private DbxClientWrapper client = null;
        private DbxUploader<FileMetadata, UploadError, UploadErrorException> uploader = null;

        public DbxOutputStream(DbxClientWrapper client, DbxUploader<FileMetadata, UploadError, UploadErrorException> uploader,
                               OutputStream out) {

            super(out);
            this.client = client;
            this.uploader = uploader;
        }

        @Override
        protected void onClose() throws IOException {
            try {
                this.uploader.finish();
            } catch (DbxException dbexc) {
                throw new IOException(dbexc);
            } finally {
                try {
                    this.uploader.close();
                } catch (Exception ignored) {
                }
                fileSystem.putClient(this.client);
            }
        }
    }

    /**
     * Provides the VFSLib instance for this filesystem.
     *
     * @return The instance
     * @since 1.6
     */
    public VFSLib getVFSLib() {
        return this.fileSystem.getVFSLib();
    }

    /**
     * Returns the attributes specific for cloud file systems (e.g. "Content-Length" for proper upload)
     *
     * @return The map
     * @since 1.6
     */
    @Override
    protected Map<String, Object> doGetAttributes() throws Exception {

        synchronized (this) {
            if (this.attributes == null) this.attributes = new HashMap<String, Object>(0);
        }
        return this.attributes;
    }

    /**
     * Sets an attribute of this cloud file.
     *
     * @param attrName The name (ATTR_CONTENT_LENGTH|...)
     * @param value    The value
     * @since 1.6
     */
    @Override
    protected void doSetAttribute(String attrName, Object value) throws Exception {

        doGetAttributes();  // Make sure cache exists
        this.attributes.put(attrName, value);
    }
}
