/*
    Copyright (c) 2005-2024 Leisenfels GmbH. All rights reserved.
    Use is subject to license terms.
*/

package com.lf.vfslib.gdrive;

import com.google.api.client.http.GenericUrl;
import com.google.api.client.http.HttpResponse;
import com.google.api.client.http.InputStreamContent;
import com.google.api.client.util.DateTime;
import com.google.api.services.drive.Drive;
import com.google.api.services.drive.model.About;
import com.google.api.services.drive.model.File;
import com.google.api.services.drive.model.FileList;
import com.google.api.services.drive.model.ParentReference;
import com.lf.vfslib.VFSLib;
import com.lf.vfslib.core.VFSLibConstants;
import com.lf.vfslib.core.VFSLibSettings;
import com.lf.vfslib.dropbox.DbxClientWrapper;
import org.apache.commons.vfs2.*;
import org.apache.commons.vfs2.provider.AbstractFileName;
import org.apache.commons.vfs2.provider.AbstractFileObject;
import org.apache.commons.vfs2.provider.UriParser;
import org.apache.commons.vfs2.util.MonitorInputStream;
import org.apache.commons.vfs2.util.MonitorOutputStream;
import org.apache.commons.vfs2.util.RandomAccessMode;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.logging.Level;


/**
 * Represents a Google Drive file or folder.
 *
 * @author Axel Schwolow
 * @created 2016-01-01
 * @since 1.6
 */
public class GDriveFileObject extends AbstractFileObject {


    /**
     * The underlying file system.
     */
    final protected GDriveFileSystem fileSystem;
    /**
     * The relative path on the Google Drive server.
     */
    protected String relPath;
    /**
     * Helper to avoid multiple refreshing.
     */
    protected boolean inRefresh = false;
    /**
     * The underlying Google Drive file/folder.
     */
    protected File googleFile = null;
    /**
     * The shared ID of the Google Drive root folder.
     */
    protected String rootID = null;
    /**
     * The attributes specific for cloud file systems (e.g. "Content-Length" for proper upload).
     */
    protected Map<String, Object> attributes = null;


    /**
     * Constructor method.
     *
     * @param name       The file name
     * @param filesystem The file system
     * @throws NullPointerException                        If a parameter is <code>null</code>
     * @throws org.apache.commons.vfs2.FileSystemException If something goes wrong
     * @since 1.6
     */
    public GDriveFileObject(AbstractFileName name, GDriveFileSystem filesystem) throws FileSystemException {

        super(name, filesystem);

        // For URL:      /Marianne Hoppe
        // For Google:   Marianne Hoppe (internally mapped)

        this.fileSystem = filesystem;
        this.relPath = UriParser.decode(filesystem.getRootName().getRelativeName(name));
        if (this.relPath.equals(".")) this.relPath = "/";  // Does not work with Google Drive!
        else if (!this.relPath.startsWith("/")) this.relPath = '/' + this.relPath;  // Files
    }

    /**
     * Clean-up method to help the gc.
     *
     * @throws Throwable Error indication
     * @since 1.6
     */
    @Override
    protected void finalize() throws Throwable {

        super.finalize();

        if (this.attributes != null) this.attributes.clear();
    }

    /**
     * Translate the absolute path for URL to a relative path compatible with Google Drive.
     *
     * @return The Google path
     * @since 1.6
     */
    private String getGooglePath() {

        if (this.relPath.equals("/")) return "";
        return this.relPath.substring(1);  // "/Marianne Hoppe" -> "Marianne Hoppe"
    }

    /**
     * Translate the absolute path for URL to a relative path compatible with Google Drive.
     *
     * @return The Google title
     * @since 1.6
     */
    private String getGoogleTitle() {

        // If relPath is "/Marianne Hoppe/Bild.jpeg" , the title is simply "Bild.jpeg"
        if (this.relPath.equals("/")) return "root";
        return this.relPath.substring(this.relPath.lastIndexOf('/') + 1);
    }

    /**
     * Checks if the current relative path denotes a file in the Google Drive root folder.
     *
     * @return Root subentry?
     * @since 1.6
     */
    private boolean isInRoot() {

        if (this.relPath.equals("/")) return true;
        return this.relPath.lastIndexOf('/') == 0;
        // "/Marianne Hoppe" is
        // "/Marianne Hoppe/Bild.jpeg" is not
    }

    /**
     * Provides the ID of the Google Drive root folder (stored separately, recycled).
     *
     * @return Root ID, <code>null</code> if not available
     * @since 1.6
     */
    private String getRootID() {

        if (this.rootID == null) {  // Read only once (static for one account)
            try {
                // TODO: May be stored for the whole filesystem? Optimize later.
                GDriveClientWrapper client = this.fileSystem.getClient();

                // Get root folder ID (stored separately)
                // Found here: https://developers.google.com/drive/v2/reference/about/get
                About about = client.driveClient.about().get().execute();
                this.rootID = about.getRootFolderId();  // 0APouqTse0V0nUk9PVA
            } catch (Exception e) {
                VFSLibSettings.log(Level.WARNING, e);
            }
        }
        return this.rootID;
    }

    /**
     * Detach from the file system.
     *
     * @since 1.6
     */
    @Override
    protected void doDetach() {
    }

    /**
     * Refresh this file.
     *
     * @throws org.apache.commons.vfs2.FileSystemException If something goes wrong
     * @since 1.6
     */
    @Override
    public void refresh() throws FileSystemException {

        if (!this.inRefresh) {  // Avoid redundant refreshing
            try {
                this.inRefresh = true;
                super.refresh();
            } finally {
                this.inRefresh = false;
            }
        }
    }

    /**
     * Determines the type of this file, returns <code>null</code> if the file does not exist.
     *
     * @return The file type
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected FileType doGetType() throws Exception {

        // Possible VFS values:
        // FileType.FILE
        // FileType.FILE_OR_FOLDER
        // FileType.FOLDER
        // FileType.IMAGINARY

        statSelf();

        // May still be null if an anticipated file is queried (e.g. abc.txt.sha)
        if (this.googleFile != null) {  // Has been resolved (exists)?

            // Special MIME type for folders
            if (this.googleFile.getMimeType().equals("application/vnd.google-apps.folder")) return FileType.FOLDER;
            else return FileType.FILE;
        }
        FileType type = FileType.IMAGINARY;
        try {
            // This is special here: IMAGINARY has no attributes, allow attributes for setting upload "Content-Length"
            Field field = type.getClass().getDeclaredField("hasAttrs");  // enum
            field.setAccessible(true);  // Override access limitations for private, protected etc.
            field.set(type, true);
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
        }
        return type;
    }

    /**
     * Determines the type of this file.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    protected void statSelf() throws Exception {

        if (this.googleFile != null) return;

        GDriveClientWrapper client = this.fileSystem.getClient();
        try {
            // Special MIME type for folders is "application/vnd.google-apps.folder"
            if (this.relPath.equals("/")) {

                Drive.Files.Get request = client.driveClient.files().get(this.getRootID());
                this.googleFile = request.execute();
            } else {
                // Get the IDs of the parent folders for subentry.
                // Google Drive does not allow to query full path like "Marianne Hoppe/Postkarte.jpeg" at once.
                String parentid = this.getRootID();  // Start from Google Drive root

                // Marianne Hoppe
                // Marianne Hoppe/subfolder
                // Marianne Hoppe/subfolder/Postkarte.jpeg

                String[] paths = this.getGooglePath().split("([/])");
                for (int i = 0; i < paths.length; i++) {
                    String next = paths[i];

                    // Get ID of next subfolder
                    Drive.Files.List request = client.driveClient.files().list();
                    request.setQ("title='" + next + "' AND trashed=false AND '" + parentid + "' in parents");

                    List<File> list = request.execute().getItems();
                    if (list.size() == 1) {
                        if (i == (paths.length - 1)) this.googleFile = list.get(0);  // Last index = desired file
                        else parentid = list.get(0).getId();
                    } else break;  // Parent does not exists
                }
            }
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e); // Needed by developers
            this.googleFile = null;  // Does not exist
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Called when the type or content of this file changes.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void onChange() throws Exception {

        this.googleFile = null;
        statSelf();
    }

    /**
     * Creates this file as a folder.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doCreateFolder() throws Exception {

        this.statSelf();

        GDriveClientWrapper client = this.fileSystem.getClient();
        try {
            // Caution: Google Drive allows creating a folder like "Marianne Hoppe/Unterordner"!

            this.doCreateEntry(client, "application/vnd.google-apps.folder");
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Gets the last modification date and time.
     *
     * @return The last modification time in milliseconds
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected long doGetLastModifiedTime() throws Exception {

        this.statSelf();

        // May be null if an anticipated file is queried (e.g. abc.txt.sha)
        try {
            DateTime modtime = this.googleFile.getModifiedDate();
            if (modtime != null) return modtime.getValue();
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
        }
        throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_UNKNOWN_MOD_TIME_ERROR"));  // Recycled
    }

    /**
     * Sets the last modified time of this file.
     *
     * @param modtime Last modification time in milliseconds
     * @return Successful?
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected boolean doSetLastModifiedTime(long modtime) throws Exception {

        this.statSelf();

        // May be null if an anticipated file is queried (e.g. abc.txt.sha)
        try {
            // TODO: Test.
            this.googleFile.setModifiedDate(new DateTime(modtime));
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
        }
        throw new FileSystemException(VFSLibSettings.getUserText(GDriveClientWrapper.class + "_SET_MOD_TIME_ERROR"));
    }

    /**
     * Deletes the file.
     * <p>
     * Entries either can be put into trash (may be restored) or be deleted permanently here.
     *
     * @throws Exception If something goes wrong
     * @see GDriveFileSystemConfigBuilder#setUseTrash(FileSystemOptions, Boolean)
     * @since 1.6
     */
    @Override
    protected void doDelete() throws Exception {

        this.statSelf();

        GDriveClientWrapper client = this.fileSystem.getClient();
        try {
            FileSystemOptions fsoptions = this.fileSystem.getFileSystemOptions();
            Boolean usetrash = GDriveFileSystemConfigBuilder.getSharedInstance().getUseTrash(fsoptions);

            if (usetrash == null || usetrash) {  // Default: use trash
                // Allow entry to be restored
                Drive.Files.Trash request = client.driveClient.files().trash(this.googleFile.getId());
                request.execute();
            } else {  // Delete permanently
                Drive.Files.Delete request = client.driveClient.files().delete(this.googleFile.getId());
                request.execute();
            }
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Rename the file.
     * <p>
     * This may take a while that Google Drive has finished renaming the file/folder internally.
     * So please give it a little timeout before <code>exists()</code> etc. is called.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doRename(FileObject newfile) throws Exception {

        this.statSelf();

        GDriveClientWrapper client = this.fileSystem.getClient();
        try {
            // Found here: https://developers.google.com/drive/v2/reference/files/patch
            File file = new File();
            file.setTitle(newfile.getName().getBaseName());

            Drive.Files.Patch request = client.driveClient.files().patch(this.googleFile.getId(), file);
            request.setFields("title");

            request.execute();

            // May take a while that exists() returns true
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Lists the children of this file.
     *
     * @return The list of children, may be empty
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected FileObject[] doListChildrenResolved() throws Exception {

        FileObject[] array = null;
        GDriveClientWrapper client = this.fileSystem.getClient();
        FileSystemManager fsmanager = getFileSystem().getFileSystemManager();


        statSelf();

        try {
            Drive.Files.List request = client.driveClient.files().list();

            if (this.relPath.equals("/")) request.setQ("'" + this.getRootID() + "' in parents AND trashed=false");
            else request.setQ("'" + this.googleFile.getId() + "' in parents AND trashed=false");

            FileList filelist = request.execute();
            List<File> list = filelist.getItems();

            int size = list.size();
            array = new FileObject[size];
            for (int i = 0; i < size; i++) {
                File child = list.get(i);
                // E.g. "/Erste Schritte.pdf"

                array[i] = this.fileSystem.resolveFile(
                        fsmanager.resolveName(getName(), UriParser.encode(child.getTitle()), NameScope.CHILD));

                // File metadata has been part of the list() response
                ((GDriveFileObject) array[i]).googleFile = child;
            }
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
            throw e;
        } finally {
            this.fileSystem.putClient(client);
        }
        return array;
    }

    /**
     * Lists the children of this file.
     *
     * @return The list of children, may be empty
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected String[] doListChildren() throws Exception {

        FileObject[] childs = this.doListChildrenResolved();
        String[] result = new String[childs.length];
        for (int i = 0; i < childs.length; i++) {
            result[i] = String.valueOf(childs[i]);
        }
        return result;
    }

    /**
     * Returns the size of the file content (in bytes).
     *
     * @return The size, -1 if not available
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected long doGetContentSize() throws Exception {

        statSelf();

        // May be null if an anticipated file is queried (e.g. abc.txt.sha)
        try {
            return this.googleFile.getFileSize();  // null if entry has no content (e.g. folder)
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
            throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_UNKNOWN_SIZE_ERROR"));  // Recycled
        }
    }

    /**
     * Provides the instance to model random access for files.
     *
     * @return The instance
     * @since 1.6
     */
    @Override
    protected RandomAccessContent doGetRandomAccessContent(RandomAccessMode mode) {
        return new GDriveRandomAccessContent(this, mode);
    }

    /**
     * Creates an input stream to read the file content from.
     *
     * @return The input stream
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected InputStream doGetInputStream() throws Exception {

        this.statSelf();

        // Only makes sense for files...
        if (!getType().hasContent()) {
            String text = VFSLibSettings.getUserText(DbxClientWrapper.class + "_READ_NOT_FILE_ERROR");  // Recycled
            throw new FileSystemException(text, getName());
        }

        GDriveClientWrapper client = this.fileSystem.getClient();

        // VFS-113: Avoid NPE
        synchronized (this.fileSystem) {
            try {
                // Found here: https://developers.google.com/drive/v2/reference/files/get#examples (Java section)
                HttpResponse response = client.driveClient.getRequestFactory().buildGetRequest(
                        new GenericUrl(this.googleFile.getDownloadUrl())).execute();

                return new GDriveInputStream(client, response.getContent());
            } finally {
                // For client release and stream close see GDriveOutputStream
            }
        }
    }

    /**
     * Creates an output stream to write the file content to.
     *
     * @param append Append? Create freshly otherwise.
     * @return The output stream, maybe <code>null</code>
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected OutputStream doGetOutputStream(boolean append) throws Exception {

        this.statSelf();

        // Only makes sense for files...
        GDriveClientWrapper client = this.fileSystem.getClient();

        // VFS-113: Avoid NPE
        synchronized (this.fileSystem) {
            try {
                // File must already exist in Google Drive (ID required for upload)
                File file = this.doCreateEntry(client, "text/plain");

                PipedOutputStream postream = new PipedOutputStream();
                PipedInputStream pistream = new PipedInputStream(postream, 4096);  // See FileSystemCopier class

                // Amazon S3 Javadoc:
                // "When uploading directly from an input stream, content length must be specified before data
                // can be uploaded to Amazon S3. If not provided, the library will have to buffer the contents
                // of the input stream in order to calculate it. Amazon S3 explicitly requires that the content
                // length be sent in the request headers before any of the data is sent."

                // Means: If not provided the whole data from the stream is cached to get the content length (JavaHeapSpace error)!

                // Content length has been set as attribute for this file?
                long len = -1;
                try {
                    // Throws FileSystemException
                    len = Long.parseLong(String.valueOf(super.getContent().getAttribute(VFSLibConstants.ATTR_CONTENT_LENGTH)));
                } catch (Exception e) {
                    VFSLibSettings.log(Level.WARNING, e);
                }

                // Send the request to the API
                InputStreamContent content = new InputStreamContent(file.getMimeType(), pistream);
                if (len != -1) content.setLength(len);
                final Drive.Files.Update request = client.driveClient.files().update(file.getId(), file, content);

                Thread thread = new Thread(() -> {
                    try {
                        request.execute();
                    } catch (Exception ignored) {
                    }
                }, VFSLibSettings.LOG_PREFIX + "Update Requester");
                thread.setPriority(Thread.MIN_PRIORITY);
                thread.start();

                return new GDriveOutputStream(client, postream, thread);  // Release client afterwards
            } catch (Exception e) {
                VFSLibSettings.log(Level.WARNING, e);
            } finally {
                // For client release and stream close see GDriveOutputStream
            }
        }
        return null;
    }

    /**
     * An <code>InputStream</code> that monitors for end-of-file.
     *
     * @since 1.6
     */
    private class GDriveInputStream extends MonitorInputStream {


        private final GDriveClientWrapper client;

        public GDriveInputStream(GDriveClientWrapper client, InputStream in) {

            super(in);
            this.client = client;
        }

        @Override
        protected void onClose() {

            // No need to close stream here
            fileSystem.putClient(this.client);
        }
    }

    /**
     * An <code>OutputStream</code> that wraps a Google Drive <code>OutputStream</code>.
     *
     * @since 1.6
     */
    private class GDriveOutputStream extends MonitorOutputStream {


        private final GDriveClientWrapper client;
        private final Thread requester;

        public GDriveOutputStream(GDriveClientWrapper client, OutputStream out, Thread requester) {

            super(out);
            this.client = client;
            this.requester = requester;
        }

        @Override
        protected void onClose() {

            try {
                // Wait for request thread to be finished, size does not fit otherwise etc.
                this.requester.join();
            } catch (Exception ignored) {
            }

            // No need to close stream here
            fileSystem.putClient(this.client);
        }
    }

    /**
     * Creates this entry (file or folder).
     *
     * @param client   The client
     * @param mimetype The MIME, use "application/vnd.google-apps.folder" for folders
     * @return The new file, <code>null</code> otherwise
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    private File doCreateEntry(GDriveClientWrapper client, String mimetype) throws Exception {

        this.statSelf();

        // Found here: https://groups.google.com/forum/#!topic/google-appengine/tr9Bqiar3ng
        File body = new File();
        body.setTitle(getGoogleTitle());
        if (mimetype != null) body.setMimeType(mimetype);

        // Need the File ID of the parent folder to create a subfolder or entry in
        List<ParentReference> parentReferences = null;
        if (this.isInRoot()) {
            parentReferences = new Vector<>(1);
            parentReferences.add(new ParentReference().setId(this.getRootID()));
        } else {  // Get subfolder of Google Drive first

            GDriveFileObject parent = (GDriveFileObject) this.getParent();

            parent.statSelf();  // Read ID etc.

            if (parent.googleFile != null) {  // Exists
                parentReferences = new Vector<>(1);
                parentReferences.add(new ParentReference().setId(parent.googleFile.getId()));
            }
        }

        // Do not create in wrong folder
        if (parentReferences != null) {
            body.setParents(parentReferences);

            Drive.Files.Insert request = client.driveClient.files().insert(body);
            return request.execute();
        }
        return null;
    }

    /**
     * Provides the VFSLib instance for this filesystem.
     *
     * @return The instance
     * @since 1.6
     */
    public VFSLib getVFSLib() {
        return this.fileSystem.getVFSLib();
    }


    /**
     * Returns the attributes specific for cloud file systems (e.g. "Content-Length" for proper upload)
     *
     * @return The map
     * @since 1.6
     */
    @Override
    protected Map<String, Object> doGetAttributes() {

        synchronized (this) {
            if (this.attributes == null) {
                this.attributes = new HashMap<>(0);
            }
        }
        return this.attributes;
    }

    /**
     * Sets an attribute of this cloud file.
     *
     * @param attrName The name (ATTR_CONTENT_LENGTH|...)
     * @param value    The value
     * @since 1.6
     */
    @Override
    protected void doSetAttribute(String attrName, Object value) throws Exception {

        doGetAttributes();  // Make sure cache exists
        this.attributes.put(attrName, value);
    }
}
