/*
    Copyright (c) 2005-2024 Leisenfels GmbH. All rights reserved.
    Use is subject to license terms.
*/

package com.lf.vfslib.s3;

import com.amazonaws.services.s3.model.*;
import com.lf.vfslib.VFSLib;
import com.lf.vfslib.core.VFSLibConstants;
import com.lf.vfslib.core.VFSLibSettings;
import com.lf.vfslib.dropbox.DbxClientWrapper;
import com.lf.vfslib.gdrive.GDriveClientWrapper;
import com.lf.vfslib.lang.JavaUtils;
import org.apache.commons.vfs2.*;
import org.apache.commons.vfs2.provider.AbstractFileName;
import org.apache.commons.vfs2.provider.AbstractFileObject;
import org.apache.commons.vfs2.provider.UriParser;
import org.apache.commons.vfs2.util.MonitorInputStream;
import org.apache.commons.vfs2.util.MonitorOutputStream;
import org.apache.commons.vfs2.util.RandomAccessMode;

import java.io.*;
import java.lang.reflect.Field;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;


/**
 * Represents a Amazon S3 file or folder.
 *
 * @author Axel Schwolow
 * @created 2016-01-01
 * @since 1.6
 */
public class S3FileObject extends AbstractFileObject {


    /**
     * The underlying file system.
     */
    final protected S3FileSystem fileSystem;
    /**
     * The relative path on the Amazon S3 server without leading slash, folders end with '/'.
     */
    protected String relPath;
    /**
     * Helper to avoid multiple refreshing.
     */
    protected boolean inRefresh = false;
    /**
     * The underlying Amazon S3 parent container.
     */
    protected Bucket s3Bucket = null;
    /**
     * The underlying Amazon S3 object, <code>null</code> for folders.
     */
    protected S3Object s3Object = null;
    /**
     * The attributes specific for cloud file systems (e.g. "Content-Length" for proper upload).
     */
    protected Map<String, Object> attributes = null;
    /**
     * Input streams are recycled by the S3 API, implement workaround for this.
     */
    protected S3ObjectInputStream s3InputStream = null;


    /**
     * Constructor method.
     *
     * @param name       The file name
     * @param filesystem The file system
     * @throws NullPointerException                        If a parameter is <code>null</code>
     * @throws org.apache.commons.vfs2.FileSystemException If something goes wrong
     * @since 1.6
     */
    public S3FileObject(AbstractFileName name, S3FileSystem filesystem) throws FileSystemException {

        super(name, filesystem);

        // Normalize files and folders to "" (root), "Marianne Hoppe" (folder), "Marianne Hoppe/Postkarte.jpeg" (file)
        this.fileSystem = filesystem;
        this.relPath = UriParser.decode(filesystem.getRootName().getRelativeName(name));
        if (this.relPath.equals(".") || this.relPath.equals("/")) this.relPath = "";  // Root (bucket)
        this.relPath = this.relPath.replaceAll("^([/]+|[/]+)$", "");

        // See statSelf() for determining if it's a file or folder, folders are added a trailing slash '/'
    }

    /**
     * Clean-up method to help the gc.
     *
     * @throws Throwable Error indication
     * @since 1.6
     */
    @Override
    protected void finalize() throws Throwable {

        super.finalize();

        if (this.attributes != null) this.attributes.clear();
    }

    /**
     * Provides the the Amazon S3 bucket (stored separately, recycled).
     *
     * @return Bucket, <code>null</code> if not available
     * @since 1.6
     */
    private Bucket getS3Bucket() {

        if (this.s3Bucket == null) {  // Read only once (static for one connection)
            try {
                FileSystemOptions fsoptions = this.fileSystem.getFileSystemOptions();
                String name = S3FileSystemConfigBuilder.getSharedInstance().getBucketName(fsoptions);
                this.s3Bucket = new Bucket(name);
            } catch (Exception ignored) {
            }
        }
        return this.s3Bucket;
    }

    /**
     * Detach from the file system.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doDetach() throws Exception {

        // Keep data as long this file object needs to be refreshed
        //this.s3Object = null;
    }

    /**
     * Refresh this file.
     *
     * @throws org.apache.commons.vfs2.FileSystemException If something goes wrong
     * @since 1.6
     */
    @Override
    public void refresh() throws FileSystemException {

        if (!this.inRefresh) {  // Avoid redundant refreshing
            try {
                this.inRefresh = true;
                super.refresh();
            } finally {
                this.inRefresh = false;
            }
        }
    }

    /**
     * Determines the type of this file, returns <code>null</code> if the file does not exist.
     *
     * @return The file type
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected FileType doGetType() throws Exception {

        // Possible VFS values:
        // FileType.FILE
        // FileType.FILE_OR_FOLDER
        // FileType.FOLDER
        // FileType.IMAGINARY

        statSelf();

        if (this.relPath.endsWith("/")) return FileType.FOLDER;
        else if (this.s3Object != null) {
            // May still be null if an anticipated file is queried (e.g. abc.txt.sha)
            return FileType.FILE;
        }
        FileType type = FileType.IMAGINARY;
        try {
            // This is special here: IMAGINARY has no attributes, allow attributes for setting upload "Content-Length"
            Field field = type.getClass().getDeclaredField("hasAttrs");  // enum
            field.setAccessible(true);  // Override access limitations for private, protected etc.
            field.set(type, true);
        } catch (Exception ignored) {
        }
        return type;
    }

    /**
     * Determines the type of this file.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    protected void statSelf() throws Exception {

        this.getS3Bucket();  // Make sure bucket here
        if (this.s3Object != null || this.relPath.endsWith("/")) return;  // File or folder

        S3ClientWrapper client = this.fileSystem.getClient();
        try {
            FileSystemOptions fsoptions = this.fileSystem.getFileSystemOptions();
            String name = S3FileSystemConfigBuilder.getSharedInstance().getBucketName(fsoptions);

            // Must be compatible with folders created with AWS console.
            // Found here: http://stackoverflow.com/questions/1939743/amazon-s3-boto-how-to-create-folder
            if (this.relPath.isEmpty()) {  // Bucket (root)
                this.relPath = "/";    // Set folder indicator
                this.s3Object = null;  // Not available here
            } else {  // Make the common case fast, assume it's a file and try to get the object reference
                this.s3Object = null;
                try {
                    this.s3Object = client.s3Client.getObject(this.s3Bucket.getName(), this.relPath); // Throws
                } catch (Exception e) {
                    //e.printStackTrace();

                    // No file, is it a folder (e.g. "Marianne Hoppe/")? List parent folder and see.
                    ListObjectsRequest request = new ListObjectsRequest();
                    request.setBucketName(this.s3Bucket.getName());
                    request.setDelimiter("/");

                    // "Marianne Hoppe/subfolder" -> "Marianne Hoppe/"
                    int pos = this.relPath.lastIndexOf('/');
                    String parentpath = pos != -1 ? this.relPath.substring(0, pos) + "/" : "";  // "" = root (bucket)
                    request.setPrefix(parentpath);  // root = "/", "Marianne Hoppe/", ...
                    String folderpath = this.relPath + "/";  // Search for subfolder

                    ObjectListing listing = client.s3Client.listObjects(request);

                    for (String folder : listing.getCommonPrefixes()) {
                        // folder=Marianne Hoppe/
                        if (folder.equals(folderpath)) {
                            this.relPath = folderpath;  // Set folder indicator
                            this.s3Object = null;  // Not available here
                            break;
                        }
                    }
                }
            }
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);  // Needed by developers
            this.s3Object = null;
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Called when the type or content of this file changes.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void onChange() throws Exception {

        //this.s3Object = null;
        statSelf();
    }

    /**
     * Creates this file as a folder.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doCreateFolder() throws Exception {

        this.statSelf();

        S3ClientWrapper client = this.fileSystem.getClient();
        try {
            // Found here: http://stackoverflow.com/questions/1939743/amazon-s3-boto-how-to-create-folder
            String folder = (this.relPath.endsWith("/") ? this.relPath : this.relPath + "/");  // Folder indicator
            PutObjectRequest request = new PutObjectRequest(this.s3Bucket.getName(), folder,
                    new ByteArrayInputStream(new byte[0]), null);
            PutObjectResult result = client.s3Client.putObject(request);

            // Remember that the S3 Management Console may need a refresh for new folders to show up
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Gets the last modification date and time.
     *
     * @return The last modification time in milliseconds
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected long doGetLastModifiedTime() throws Exception {

        this.statSelf();

        // May be null if an anticipated file is queried (e.g. abc.txt.sha)
        try {
            return this.s3Object.getObjectMetadata().getLastModified().getTime();
            // TODO: Not set for folders. Other way to get info?
        } catch (Exception e) {
            throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_UNKNOWN_MOD_TIME_ERROR"));  // Recycled
        }
    }

    /**
     * Sets the last modified time of this file (not supported by Amazon S3).
     *
     * @param modtime Last modification time in milliseconds
     * @return Successful?
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected boolean doSetLastModifiedTime(long modtime) throws Exception {

        this.statSelf();

        // May be null if an anticipated file is queried (e.g. abc.txt.sha)
        try {
            // TODO: Test.
            this.s3Object.getObjectMetadata().setLastModified(new Date(modtime));  // Throws
        } catch (Exception e) {
            //VFSLibSettings.log(Level.WARNING, e);
        }
        throw new FileSystemException(VFSLibSettings.getUserText(GDriveClientWrapper.class + "_SET_MOD_TIME_ERROR"));  // Recycled
    }

    /**
     * Deletes the file.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doDelete() throws Exception {

        this.statSelf();

        S3ClientWrapper client = this.fileSystem.getClient();
        try {
            // Works also for folders
            client.s3Client.deleteObject(this.s3Bucket.getName(), this.relPath);
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Rename the file.
     * <p>
     * Renaming is not currently supported by Amazon S3, so let's do the copy-and-delete thing for files.
     * Remember that renaming folders is currently not possible for Amazon S3 at all. The S3 Management
     * Console does not support it and neither does VFSLib.
     *
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected void doRename(FileObject newfile) throws Exception {

        this.statSelf();

        S3ClientWrapper client = this.fileSystem.getClient();
        try {
            // Found here:
            // http://stackoverflow.com/questions/2481685/amazon-s3-boto-how-to-rename-a-file-in-a-bucket
            // http://stackoverflow.com/questions/8396224/how-to-rename-a-folder-inside-a-bucket-using-the-amazons3-net-sdk

            if (this.getType().equals(FileType.FILE)) {
                // Copy and delete source
                client.s3Client.copyObject(this.s3Bucket.getName(), this.relPath,                       // Source
                        this.s3Bucket.getName(), ((S3FileObject) newfile).relPath);  // Target

                client.s3Client.deleteObject(this.s3Bucket.getName(), this.relPath);
            } else if (this.getType().equals(FileType.FOLDER)) {
                // TODO: No way to do it right now. Recursive copying seems very laborious. Wait for Amazon solution.
                // TODO: Remember that renaming a folder is not supported by the S3 Management Console as well.
            } else {
                throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_RENAME_ERROR"));  // Recycled
            }
        } finally {
            this.fileSystem.putClient(client);
        }
    }

    /**
     * Lists the children of this file.
     *
     * @return The list of children, may be empty
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected FileObject[] doListChildrenResolved() throws Exception {

        FileObject[] array = null;
        S3ClientWrapper client = this.fileSystem.getClient();
        FileSystemManager fsmanager = getFileSystem().getFileSystemManager();


        statSelf();

        try {
            // Must be compatible with folders created with AWS console.
            // Found here: http://stackoverflow.com/questions/1939743/amazon-s3-boto-how-to-create-folder
            ListObjectsRequest request = new ListObjectsRequest();
            request.setBucketName(this.s3Bucket.getName());
            request.setDelimiter("/");
            request.setPrefix(this.relPath.equals("/") ? "" : this.relPath);  // root = "", "Marianne Hoppe/", ...

            ObjectListing listing = client.s3Client.listObjects(request);

            // 1. step: Get folders (prefixes here)
            java.util.List<String> folders = listing.getCommonPrefixes();
            int size = folders.size();
            array = new FileObject[size];
            for (int i = 0; i < size; i++) {
                String folder = folders.get(i);

                String[] list = folder.split("([/])");  // Need only last part
                folder = list[list.length - 1];

                array[i] = this.fileSystem.resolveFile(
                        fsmanager.resolveName(getName(), UriParser.encode(folder), NameScope.CHILD));

                // No statSelf() necessary, must be a folder
                if (!((S3FileObject) array[i]).relPath.endsWith("/")) {
                    ((S3FileObject) array[i]).relPath += "/";  // Set folder indicator
                }
            }

            // 2. step: Get files (objects here)
            java.util.List<S3ObjectSummary> children = listing.getObjectSummaries();
            size = children.size();
            for (int i = 0; i < size; i++) {
                S3ObjectSummary summary = children.get(i);

                // Caution: file list contains the folder itself (seems to be a bug)
                if (summary.getKey().equals(this.relPath)) continue;

                String file = summary.getKey();  // "Marianne Hoppe/Postkarte.jpeg"
                String[] list = file.split("([/])");  // Need only last part
                file = list[list.length - 1];

                FileObject fileobj = this.fileSystem.resolveFile(
                        fsmanager.resolveName(getName(), UriParser.encode(file), NameScope.CHILD));
                array = (FileObject[]) JavaUtils.addToArray(array, fileobj);

                // Call statSelf() here
            }
        } catch (Exception e) {
            VFSLibSettings.log(Level.WARNING, e);
            throw e;
        } finally {
            this.fileSystem.putClient(client);
        }
        return array;
    }

    /**
     * Lists the children of this file.
     *
     * @return The list of children, may be empty
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected String[] doListChildren() throws Exception {

        FileObject[] childs = this.doListChildrenResolved();
        String[] result = new String[childs.length];
        for (int i = 0; i < childs.length; i++) {
            result[i] = String.valueOf(childs[i]);
        }
        return result;
    }

    /**
     * Returns the size of the file content (in bytes).
     *
     * @return The size, -1 if not available
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected long doGetContentSize() throws Exception {

        statSelf();

        // May be null if an anticipated file is queried (e.g. abc.txt.sha)
        try {
            return this.s3Object.getObjectMetadata().getContentLength();
        } catch (Exception e) {
            throw new FileSystemException(VFSLibSettings.getUserText(DbxClientWrapper.class + "_UNKNOWN_SIZE_ERROR"));  // Recycled
        }
    }

    /**
     * Provides the instance to model random access for files.
     *
     * @return The instance
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected RandomAccessContent doGetRandomAccessContent(RandomAccessMode mode) throws Exception {
        return new S3RandomAccessContent(this, mode);
    }

    /**
     * Creates an input stream to read the file content from.
     *
     * @return The input stream
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected InputStream doGetInputStream() throws Exception {

        this.statSelf();

        // Only makes sense for files...
        if (!getType().hasContent()) {
            String text = VFSLibSettings.getUserText(DbxClientWrapper.class + "_READ_NOT_FILE_ERROR");  // Recycled
            throw new FileSystemException(text, getName());
        }

        S3ClientWrapper client = this.fileSystem.getClient();

        // VFS-113: Avoid NPE
        synchronized (this.fileSystem) {
            try {
                // Problem: Object input stream can only be requested once. If the stream has been closed
                // then the second call will fail (calculate-only digests for MD5 and SHA-1 with two tasks).
                S3ObjectInputStream istream = this.s3Object.getObjectContent();
                if (istream == this.s3InputStream) {
                    // Workaround: create file object clone with different stream
                    S3Object clone = client.s3Client.getObject(this.s3Bucket.getName(), this.relPath); // Throws
                    istream = clone.getObjectContent();
                } else this.s3InputStream = istream;  // Remember
                return new S3InputStream(client, istream);
            } finally {
                // For client release and stream close see S3InputStream
            }
        }
    }

    /**
     * Creates an output stream to write the file content to.
     *
     * @param append Not supported currently, always created freshly
     * @return The output stream
     * @throws Exception If something goes wrong
     * @since 1.6
     */
    @Override
    protected OutputStream doGetOutputStream(boolean append) throws Exception {

        this.statSelf();

        // Only makes sense for files...
        //if (!getType().hasContent()) throw new FileSystemException("vfs.provider/write-not-file.error", getName());

        final S3ClientWrapper client = this.fileSystem.getClient();

        // VFS-113: Avoid NPE
        synchronized (this.fileSystem) {
            try {
                // TODO: Any way to support append?

                PipedOutputStream postream = new PipedOutputStream();
                PipedInputStream pistream = new PipedInputStream(postream, 4096);  // See FileSystemCopier class

                // Amazon S3 Javadoc:
                // "When uploading directly from an input stream, content length must be specified before data
                // can be uploaded to Amazon S3. If not provided, the library will have to buffer the contents
                // of the input stream in order to calculate it. Amazon S3 explicitly requires that the content
                // length be sent in the request headers before any of the data is sent."

                // Means: If not provided the whole data from the stream is cached to get the content length (JavaHeapSpace error)!
                ObjectMetadata metadata = null;

                // Content length has been set as attribute for this file?
                try {
                    // Throws FileSystemException
                    long len = Long.parseLong(String.valueOf(super.getContent().getAttribute(VFSLibConstants.ATTR_CONTENT_LENGTH)));
                    metadata = new ObjectMetadata();
                    metadata.setContentLength(len);
                } catch (Exception ignored) {
                }

                final PutObjectRequest request = new PutObjectRequest(this.s3Bucket.getName(), this.relPath, pistream, metadata);

                Thread thread = new Thread(() -> {
                    try {
                        PutObjectResult result = client.s3Client.putObject(request);
                    } catch (Exception ignored) {
                    }
                }, VFSLibSettings.LOG_PREFIX + "Put Object Requester");
                thread.setPriority(Thread.MIN_PRIORITY);
                thread.start();

                return new S3OutputStream(client, postream, thread);  // Release client afterwards
            } finally {
                // For client release and stream close see S3OutputStream
            }
        }
    }

    /**
     * An <code>InputStream</code> that monitors for end-of-file.
     *
     * @since 1.6
     */
    private class S3InputStream extends MonitorInputStream {

        private S3ClientWrapper client = null;

        public S3InputStream(S3ClientWrapper client, InputStream in) {

            super(in);
            this.client = client;
        }

        @Override
        protected void onClose() {

            // No need to close stream here
            fileSystem.putClient(this.client);
        }
    }

    /**
     * An <code>OutputStream</code> that wraps a Amazon S3 <code>OutputStream</code>.
     *
     * @since 1.6
     */
    private class S3OutputStream extends MonitorOutputStream {

        private S3ClientWrapper client = null;
        private Thread requester = null;

        public S3OutputStream(S3ClientWrapper client, OutputStream out, Thread requester) {

            super(out);
            this.client = client;
            this.requester = requester;
        }

        @Override
        protected void onClose() {

            try {
                // Wait for request thread to be finished, size does not fit otherwise etc.
                this.requester.join();
            } catch (Exception ignored) {
            }

            // No need to close stream here
            fileSystem.putClient(this.client);
        }
    }

    /**
     * Provides the VFSLib instance for this filesystem.
     *
     * @return The instance
     * @since 1.6
     */
    public VFSLib getVFSLib() {
        return this.fileSystem.getVFSLib();
    }


    /**
     * Returns the attributes specific for cloud file systems (e.g. "Content-Length" for proper upload)
     *
     * @return The map
     * @since 1.6
     */
    @Override
    protected Map<String, Object> doGetAttributes() {

        synchronized (this) {
            if (this.attributes == null) this.attributes = new HashMap<>(0);
        }
        return this.attributes;
    }

    /**
     * Sets an attribute of this cloud file.
     *
     * @param attrName The name (ATTR_CONTENT_LENGTH|...)
     * @param value    The value
     * @since 1.6
     */
    @Override
    protected void doSetAttribute(String attrName, Object value) {

        doGetAttributes();  // Make sure cache exists
        this.attributes.put(attrName, value);
    }
}
