/*
    Copyright (c) 2005-2024 Leisenfels GmbH. All rights reserved.
    Use is subject to license terms.
*/

package com.lf.vfslib.s3;

import com.lf.vfslib.VFSLib;
import org.apache.commons.vfs2.FileSystem;
import org.apache.commons.vfs2.FileSystemConfigBuilder;
import org.apache.commons.vfs2.FileSystemOptions;


/**
 * The config builder for the various Amazon S3 configuration options.
 *
 * @author Axel Schwolow
 * @created 2016-01-01
 * @since 1.6
 */
public class S3FileSystemConfigBuilder extends FileSystemConfigBuilder {


    /**
     * Shared <code>DbxFileSystemConfigBuilder</code> instance.
     */
    protected static S3FileSystemConfigBuilder sharedInstance;

    /**
     * Variable name.
     */
    public final static String DISPLAY_NAME = S3FileSystemConfigBuilder.class.getName() + ".DISPLAY_NAME";
    /**
     * Variable name.
     */
    public final static String ACCESS_KEY_ID = S3FileSystemConfigBuilder.class.getName() + ".ACCESS_KEY_ID";
    /**
     * Variable name.
     */
    public final static String SECRET_KEY = S3FileSystemConfigBuilder.class.getName() + ".SECRET_KEY";
    /**
     * Variable name.
     */
    public final static String BUCKET_NAME = S3FileSystemConfigBuilder.class.getName() + ".BUCKET_NAME";
    /**
     * Variable name.
     */
    public final static String BUCKET_URI = S3FileSystemConfigBuilder.class.getName() + ".BUCKET_URI";

    /**
     * The parental object.
     */
    protected VFSLib vfsLib;


    /**
     * Constructor method.
     * <p>
     * Uses the shared instance of the <code>VFSLib</code> class and its resource bundle.
     *
     * @since 1.6
     */
    public S3FileSystemConfigBuilder() {
        this(null);
    }

    /**
     * Constructor method.
     *
     * @param vfslib The parental object, shared instance if <code>null</code>
     * @since 1.6
     */
    public S3FileSystemConfigBuilder(VFSLib vfslib) {

        super("s3.");  // Static to find resources, do not use dynamic scheme here
        this.vfsLib = (vfslib != null ? vfslib : VFSLib.getSharedInstance());
    }

    /**
     * Provides the shared instance of the factory implementation.
     *
     * @return Shared instance
     * @since 1.6
     */
    public static synchronized S3FileSystemConfigBuilder getSharedInstance() {

        if (sharedInstance == null) sharedInstance = new S3FileSystemConfigBuilder();
        return sharedInstance;
    }

    /**
     * Provides the file system class to be configured by this builder.
     *
     * @return File system class
     * @since 1.6
     */
    @Override
    protected Class<? extends FileSystem> getConfigClass() {
        return S3FileSystem.class;
    }

    /**
     * Setter for Amazon S3 account display name.
     *
     * @param options The options
     * @param name    The display name
     * @since 1.6
     */
    public void setAccountDisplayName(FileSystemOptions options, String name) {
        super.setParam(options, DISPLAY_NAME, name);
    }

    /**
     * Getter for Amazon S3 account display name.
     *
     * @param options The options
     * @return The display name, may be <code>null</code>
     * @since 1.6
     */
    public String getAccountDisplayName(FileSystemOptions options) {
        return super.getParam(options, DISPLAY_NAME);
    }

    /**
     * Setter for Amazon S3 access key ID.
     *
     * @param options The options
     * @param token   The access key ID
     * @since 1.6
     */
    public void setAccessKeyID(FileSystemOptions options, String token) {
        super.setParam(options, ACCESS_KEY_ID, token);
    }

    /**
     * Getter for Amazon S3 access key ID.
     *
     * @param options The options
     * @return The access key ID, may be <code>null</code>
     * @since 1.6
     */
    public String getAccessKeyID(FileSystemOptions options) {
        return super.getParam(options, ACCESS_KEY_ID);
    }

    /**
     * Setter for Amazon S3 secret key.
     *
     * @param options The options
     * @param token   The secret key
     * @since 1.6
     */
    public void setSecretKey(FileSystemOptions options, String token) {
        super.setParam(options, SECRET_KEY, token);
    }

    /**
     * Getter for Amazon S3 secret key.
     *
     * @param options The options
     * @return The secret key, may be <code>null</code>
     * @since 1.6
     */
    public String getSecretKey(FileSystemOptions options) {
        return super.getParam(options, SECRET_KEY);
    }

    /**
     * Setter for the name of the Amazon S3 bucket.
     *
     * @param options The options
     * @param name    The display name
     * @since 1.6
     */
    public void setBucketName(FileSystemOptions options, String name) {
        super.setParam(options, BUCKET_NAME, name);
    }

    /**
     * Getter for the name of the Amazon S3 bucket.
     *
     * @param options The options
     * @return The display name, may be <code>null</code>
     * @since 1.6
     */
    public String getBucketName(FileSystemOptions options) {
        return super.getParam(options, BUCKET_NAME);
    }

    /**
     * Setter for URI mode: use bucket name or host.
     *
     * @param options The options
     * @param enable  Use bucket name?
     * @since 1.6
     */
    public void setUseBucketURI(FileSystemOptions options, Boolean enable) {
        super.setParam(options, BUCKET_URI, enable);
    }

    /**
     * Getter for URI mode: use bucket name or host.
     *
     * @param options The options
     * @return Use bucket name? May be <code>null</code>.
     * @since 1.6
     */
    public Boolean getUseBucketURI(FileSystemOptions options) {
        return super.getParam(options, BUCKET_URI);
    }
}
